from httpx import HTTPError, HTTPStatusError


class AnaplanException(Exception):
    """
    Base class for all Anaplan SDK Exceptions.
    """

    def __init__(self, message: str):
        self.message = f"\n\n\n{message}\n"
        super().__init__(self.message)


class InvalidCredentialsException(AnaplanException):
    """
    Exception raised when the provided credentials are invalid.
    """

    def __init__(self, message: str = "Invalid credentials."):
        self.message = message
        super().__init__(self.message)


class InvalidPrivateKeyException(InvalidCredentialsException):
    """
    Exception raised when the provided private key is invalid.
    """

    def __init__(self, message: str = "Invalid private key."):
        self.message = message
        super().__init__(self.message)


class InvalidIdentifierException(AnaplanException):
    """
    Exception raised when the provided identifier is invalid.
    """

    def __init__(self, message: str = "Invalid identifier."):
        self.message = message
        super().__init__(self.message)


class AnaplanActionError(AnaplanException):
    """
    Exception raised when an Anaplan Action fails.
    """

    def __init__(self, message: str = "Anaplan completed with errors."):
        self.message = message
        super().__init__(self.message)


def raise_appropriate_error(error: HTTPError) -> None:
    if isinstance(error, HTTPStatusError):
        if error.response.status_code == 404:
            raise InvalidIdentifierException from error
    raise error



"""
Custom Authentication class to pass to httpx alongside some helper functions.
"""

import base64
import logging
import os

import httpx
from cryptography.exceptions import InvalidKey, UnsupportedAlgorithm
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from ._exceptions import InvalidPrivateKeyException, InvalidCredentialsException

logger = logging.getLogger("anaplan_sdk")


class AnaplanBasicAuth(httpx.Auth):
    requires_response_body = True

    def __init__(
        self,
        user_email: str,
        password: str,
    ):
        self.token = "none"
        self.user_email = user_email
        self.password = password

    def auth_flow(self, request):
        request.headers["Authorization"] = f"AnaplanAuthToken {self.token}"
        response = yield request
        if response.status_code == 401:
            logger.info("Token expired, refreshing.")
            response = yield self._basic_auth_request()
            if "tokenInfo" not in response.json():
                raise InvalidCredentialsException
            self.token = response.json().get("tokenInfo").get("tokenValue")
            request.headers["Authorization"] = f"AnaplanAuthToken {self.token}"
            yield request

    def _basic_auth_request(self):
        credentials = base64.b64encode(f"{self.user_email}:{self.password}".encode()).decode()
        return httpx.Request(
            method="post",
            url="https://auth.anaplan.com/token/authenticate",
            headers={"Authorization": f"Basic {credentials}"},
        )


class AnaplanCertAuth(httpx.Auth):
    requires_response_body = True
    requires_request_body = True

    def __init__(
        self,
        certificate: bytes,
        private_key: RSAPrivateKey,
    ):
        self.token = "none"
        self.certificate = certificate
        self.private_key = private_key

    def auth_flow(self, request):
        request.headers["Authorization"] = f"AnaplanAuthToken {self.token}"
        response = yield request
        if response.status_code == 401:
            logger.info("Token expired, refreshing.")
            response = yield self._cert_auth_request()
            if "tokenInfo" not in response.json():
                raise InvalidCredentialsException
            self.token = response.json().get("tokenInfo").get("tokenValue")
            request.headers["Authorization"] = f"AnaplanAuthToken {self.token}"
            yield request

    def _cert_auth_request(self):
        message = os.urandom(150)
        encoded_cert = base64.b64encode(self.certificate).decode()
        encoded_string = base64.b64encode(message).decode()
        encoded_signed_string = base64.b64encode(
            self.private_key.sign(message, padding.PKCS1v15(), hashes.SHA512())
        ).decode()
        return httpx.Request(
            method="post",
            url="https://auth.anaplan.com/token/authenticate",
            headers={
                "Authorization": f"CACertificate {encoded_cert}",
                "Content-Type": "application/json",
            },
            json={"encodedData": encoded_string, "encodedSignedData": encoded_signed_string},
        )


def get_certificate(certificate: str | bytes) -> bytes:
    """
    Get the certificate from a file or string.
    :param certificate: The certificate to use.
    :return: The certificate as bytes.
    """
    if isinstance(certificate, str):
        if os.path.isfile(certificate):
            with open(certificate, "rb") as f:
                return f.read()
        return certificate.encode()
    return certificate


def get_private_key(private_key: str | bytes, private_key_password: str | bytes) -> RSAPrivateKey:
    """
    Get the private key from a file or string.
    :param private_key: The private key to use.
    :param private_key_password: The password for the private key.
    :return: The private key as an RSAPrivateKey object.
    """
    try:
        if isinstance(private_key, str):
            if os.path.isfile(private_key):
                with open(private_key, "rb") as f:
                    data = f.read()
            else:
                data = private_key.encode()
        else:
            data = private_key

        password = None
        if private_key_password:
            if isinstance(private_key_password, str):
                password = private_key_password.encode()
            else:
                password = private_key_password
        return serialization.load_pem_private_key(data, password, backend=default_backend())
    except (IOError, InvalidKey, UnsupportedAlgorithm) as error:
        raise InvalidPrivateKeyException from error



"""
Synchronous Client.
"""

import gzip
import logging
import time
from concurrent.futures import ThreadPoolExecutor

import httpx
from httpx import HTTPError

from ._auth import AnaplanBasicAuth, AnaplanCertAuth, get_certificate, get_private_key
from ._exceptions import (
    AnaplanActionError,
    raise_appropriate_error,
)
from ._models import (
    Import,
    Export,
    Process,
    File,
    Action,
    List,
    Workspace,
    Model,
    to_workspaces,
    to_models,
    to_actions,
    to_imports,
    to_exports,
    to_processes,
    to_files,
    to_lists,
    determine_action_type,
)

logger = logging.getLogger("anaplan_sdk")


class Client:
    """
    A synchronous Client for pythonic access to the Anaplan Integration API v2:
    https://anaplan.docs.apiary.io/. This Client provides high-level abstractions over the API, so
    you can deal with python objects and simple functions rather than implementation details like
    http, json, compression, chunking etc.


    For more information, quick start guides and detailed instructions refer to:
    https://vinzenzklass.github.io/anaplan-sdk.
    """

    def __init__(
        self,
        workspace_id: str | None = None,
        model_id: str | None = None,
        user_email: str | None = None,
        password: str | None = None,
        certificate: str | bytes | None = None,
        private_key: str | bytes | None = None,
        private_key_password: str | bytes | None = None,
        timeout: int = 30,
        status_poll_delay: int = 1,
        upload_parallel: bool = True,
        upload_chunk_size: int = 25_000_000,
    ) -> None:
        """
        A synchronous Client for pythonic access to the Anaplan Integration API v2:
        https://anaplan.docs.apiary.io/. This Client provides high-level abstractions over the API,
        so you can deal with python objects and simple functions rather than implementation details
        like http, json, compression, chunking etc.


        For more information, quick start guides and detailed instructions refer to:
        https://vinzenzklass.github.io/anaplan-sdk.

        :param workspace_id: The Anaplan workspace Id. You can copy this from the browser URL or
                             find them using an HTTP Client like Postman, Paw, Insomnia etc.
        :param model_id: The identifier of the model.
        :param user_email: A valid email registered with the Anaplan Workspace you are attempting
                           to access. **The associated user must have Workspace Admin privileges**
        :param password: Password for the given `user_email`. This is not suitable for production
                         setups. If you intend to use this in production, acquire a client
                         certificate as described under: https://help.anaplan.com/procure-ca-certificates-47842267-2cb3-4e38-90bf-13b1632bcd44
        :param certificate: The absolute path to the client certificate file or the certificate
                            itself.
        :param private_key: The absolute path to the private key file or the private key itself.
        :param private_key_password: The password to access the private key if there is one.
        :param timeout: The timeout for the HTTP requests.
        :param status_poll_delay: The delay between polling the status of a task.
        :param upload_parallel: Whether to upload the chunks in parallel. Defaults to True. **If
                                you are heavily network bound or are experiencing rate limiting
                                issues, set this to False.**
        :param upload_chunk_size: The size of the chunks to upload. This is the maximum size of
                                  each chunk. Defaults to 25MB.
        """
        if not ((user_email and password) or (certificate and private_key)):
            raise ValueError(
                "Either `certificate` and `private_key` or `user_email` and `password` must be "
                "provided."
            )
        auth = (
            AnaplanCertAuth(
                get_certificate(certificate), get_private_key(private_key, private_key_password)
            )
            if certificate
            else AnaplanBasicAuth(user_email=user_email, password=password)
        )
        self._client = httpx.Client(auth=auth)
        self._base_url = "https://api.anaplan.com/2/0/workspaces"
        self.workspace_id = workspace_id
        self.model_id = model_id
        self.timeout = timeout
        self.status_poll_delay = status_poll_delay
        self.upload_parallel = upload_parallel
        self.upload_chunk_size = upload_chunk_size

    def list_workspaces(self) -> list[Workspace]:
        """
        Lists all the Workspaces the authenticated user has access to.
        :return: All Workspaces as a list of :py:class:`Workspace`.
        """
        return to_workspaces(self._get(f"{self._base_url}?tenantDetails=true"))

    def list_models(self) -> list[Model]:
        """
        Lists all the Models the authenticated user has access to.
        :return: All Models in the Workspace as a list of :py:class:`Model`.
        """
        return to_models(
            self._get(f"{self._base_url.replace('/workspaces', '/models')}?modelDetails=true")
        )

    def list_actions(self) -> list[Action]:
        """
        Lists all the Actions in the Model. This will only return the Actions listed under
        `Other Actions` in Anaplan. For Imports, exports, and processes, see their respective
        methods instead.

        :return: All Actions on this model as a list of :py:class:`Action`.
        """
        return to_actions(
            self._get(f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/actions")
        )

    def list_imports(self) -> list[Import]:
        """
        Lists all the Imports in the Model.
        :return: All Imports on this model as a list of :py:class:`Import`.
        """
        return to_imports(
            self._get(f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/imports")
        )

    def list_exports(self) -> list[Export]:
        """
        Lists all the Exports in the Model.
        :return: All Exports on this model as a list of :py:class:`Export`.
        """
        return to_exports(
            self._get(f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/exports")
        )

    def list_processes(self) -> list[Process]:
        """
        Lists all the Processes in the Model.
        :return: All Processes on this model as a list of :py:class:`Process`.
        """
        return to_processes(
            self._get(f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/processes")
        )

    def list_files(self) -> list[File]:
        """
        Lists all the Files in the Model.
        :return: All Files on this model as a list of :py:class:`File`.
        """
        return to_files(
            self._get(f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/files")
        )

    def list_lists(self) -> list[List]:
        """
        Lists all the Lists in the Model.
        :return: All Lists on this model as a list of :py:class:`List`.
        """
        return to_lists(
            self._get(f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/lists")
        )

    def run_action(self, action_id: int) -> None:
        """
        Runs the specified Anaplan Action and validates the spawned task. If the Action fails or
        completes with errors, will raise an :py:class:`AnaplanActionError`. Failed Tasks are
        usually not something you can recover from at runtime and often require manual changes in
        Anaplan, i.e. updating the mapping of an Import or similar. So, for convenience, this will
        raise an Exception to handle - if you for e.g. think that one of the uploaded chunks may
        have been dropped and simply retrying with new data may help - and not return the task
        status information that needs to be handled by the caller.

        If you need more information or control, you can use `invoke_action()` and
        `get_task_status()`.
        :param action_id: The identifier of the Action to run. Can be any Anaplan Invokable;
                          Processes, Imports, Exports, Other Actions.
        """
        task_id = self.invoke_action(action_id)
        task_status = self.get_task_status(action_id, task_id)

        while "COMPLETE" not in task_status.get("taskState"):
            time.sleep(self.status_poll_delay)
            task_status = self.get_task_status(action_id, task_id)

        if task_status.get("taskState") == "COMPLETE" and not task_status.get("result").get(
            "successful"
        ):
            raise AnaplanActionError(f"Task '{task_id}' completed with errors.")

    def get_file(self, file_id: int) -> bytes:
        """
        Retrieves the content of the specified file.
        :param file_id: The identifier of the file to retrieve.
        :return: The content of the file.
        """
        return self._get_binary(
            f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/files/{file_id}"
        )

    def upload_file(self, file_id: int, content: str | bytes) -> None:
        """
        Uploads the content to the specified file. If `upload_parallel` is set to True on the
        instance you are invoking this from, will attempt to upload the chunks in parallel for
        better performance. If you are network bound or are experiencing rate limiting issues, set
        `upload_parallel` to False.

        :param file_id: The identifier of the file to upload to.
        :param content: The content to upload. **This Content will be compressed before uploading.
                        If you are passing the Input as bytes, pass it uncompressed to avoid
                        redundant work.**
        """
        if isinstance(content, str):
            content = content.encode()
        chunks = [
            content[i : i + self.upload_chunk_size]
            for i in range(0, len(content), self.upload_chunk_size)
        ]
        self._set_chunk_count(file_id, len(chunks))
        if self.upload_parallel:
            with ThreadPoolExecutor(max_workers=4) as executor:
                executor.map(
                    self._upload_chunk, (file_id,) * len(chunks), range(len(chunks)), chunks
                )
        else:
            for index, chunk in enumerate(chunks):
                self._upload_chunk(file_id, index, chunk)
        logger.info(f"Content loaded to  File '{file_id}'.")

    def get_task_status(
        self, action_id: int, task_id: str
    ) -> dict[str, float | int | str | list | dict | bool]:
        """
        Retrieves the status of the specified task.
        :param action_id: The identifier of the action that was invoked.
        :param task_id: The identifier of the spawned task.
        :return: The status of the task as returned by the API. For more information
                 see: https://anaplan.docs.apiary.io.
        """
        return self._get(
            f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/"
            f"{determine_action_type(action_id)}/{action_id}/tasks/{task_id}"
        ).get("task")

    def invoke_action(self, action_id: int) -> str:
        """
        You may want to consider using `run_action()` instead.

        Invokes the specified Anaplan Action and returns the spawned Task identifier. This is
        useful if you want to handle the Task status yourself or if you want to run multiple
        Actions in parallel.
        :param action_id:
        :return:
        """
        response = self._post(
            f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/"
            f"{determine_action_type(action_id)}/{action_id}/tasks",
            json={"localeName": "en_US"},
        )
        task_id = response.get("task").get("taskId")
        logger.info(f"Invoked Action '{action_id}', spawned Task: '{task_id}'.")
        return task_id

    def _upload_chunk(self, file_id: int, index: int, chunk: bytes) -> None:
        try:
            response = self._client.put(
                f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/files/{file_id}/"
                f"chunks/{index}",
                headers={"Content-Type": "application/x-gzip"},
                content=gzip.compress(chunk),
                timeout=self.timeout,
            )
            response.raise_for_status()
        except HTTPError as error:
            raise_appropriate_error(error)

    def _set_chunk_count(self, file_id: int, num_chunks: int) -> None:
        self._post(
            f"{self._base_url}/{self.workspace_id}/models/{self.model_id}/files/{file_id}",
            json={"chunkCount": num_chunks},
        )

    def _get(self, url: str) -> dict[str, float | int | str | list | dict | bool]:
        try:
            response = self._client.get(url, timeout=self.timeout)
            response.raise_for_status()
            return response.json()
        except HTTPError as error:
            raise_appropriate_error(error)

    def _get_binary(self, url: str) -> bytes:
        try:
            response = self._client.get(url, timeout=self.timeout)
            response.raise_for_status()
            return response.content
        except HTTPError as error:
            raise_appropriate_error(error)

    def _post(
        self, url: str, json: dict | None = None
    ) -> dict[str, float | int | str | list | dict | bool]:
        try:
            return self._client.post(
                url,
                headers={"Content-Type": "application/json"},
                json=json,
                timeout=self.timeout,
            ).json()
        except HTTPError as error:
            raise_appropriate_error(error)



from dataclasses import dataclass

from ._exceptions import InvalidIdentifierException


@dataclass
class Import:
    """
    Object representing an Anaplan Import.
    """

    id: int
    name: str
    type: str
    source_id: int | None


@dataclass
class Export:
    """
    Object representing an Anaplan Export.
    """

    id: int
    name: str
    type: str
    format: str
    encoding: str
    layout: str


@dataclass
class Action:
    """
    Object representing an Anaplan Action.
    """

    id: int
    name: str
    type: str


@dataclass
class Process:
    """
    Object representing an Anaplan Process.
    """

    id: int
    name: str


@dataclass
class File:
    """
    Object representing an Anaplan File.
    """

    id: int
    name: str
    chunk_count: int
    delimiter: str
    encoding: str
    first_data_row: int
    format: str
    header_row: int
    separator: str


@dataclass
class List:
    """
    Object representing an Anaplan List.
    """

    id: int
    name: str


@dataclass
class Workspace:
    """
    Object representing an Anaplan Workspace.
    """

    id: str
    name: str
    active: str
    size_allowance: int
    current_size: int


@dataclass
class Model:
    """
    Object representing an Anaplan Model.
    """

    id: str
    name: str
    active_state: str
    last_saved_serial_number: int
    last_modified_by_user_guid: str
    memory_usage: int
    current_workspace_id: str
    current_workspace_name: str
    model_url: str
    category_values: list
    iso_creation_date: str
    last_modified: str


def to_imports(response: dict[str, float | int | str | list | dict | bool]) -> list[Import]:
    return [
        Import(
            id=int(e.get("id")),
            type=e.get("importType"),
            name=e.get("name"),
            source_id=int(e.get("importDataSourceId")) if e.get("importDataSourceId") else None,
        )
        for e in response.get("imports")
    ]


def to_exports(response: dict[str, float | int | str | list | dict | bool]) -> list[Export]:
    return [
        Export(
            id=int(e.get("id")),
            name=e.get("name"),
            type=e.get("exportType"),
            format=e.get("exportFormat"),
            encoding=e.get("encoding"),
            layout=e.get("layout"),
        )
        for e in response.get("exports")
    ]


def to_actions(response: dict[str, float | int | str | list | dict | bool]) -> list[Action]:
    return [
        Action(id=int(e.get("id")), name=e.get("name"), type=e.get("actionType"))
        for e in response.get("actions")
    ]


def to_processes(response: dict[str, float | int | str | list | dict | bool]) -> list[Process]:
    return [Process(id=int(e.get("id")), name=e.get("name")) for e in response.get("processes")]


def to_files(response: dict[str, float | int | str | list | dict | bool]) -> list[File]:
    return [
        File(
            id=int(e.get("id")),
            name=e.get("name"),
            chunk_count=e.get("chunkCount"),
            delimiter=e.get("delimiter"),
            encoding=e.get("encoding"),
            first_data_row=e.get("firstDataRow"),
            format=e.get("format"),
            header_row=e.get("headerRow"),
            separator=e.get("separator"),
        )
        for e in response.get("files")
    ]


def to_lists(response: dict[str, float | int | str | list | dict | bool]) -> list[List]:
    return [List(id=int(e.get("id")), name=e.get("name")) for e in response.get("lists")]


def to_workspaces(response: dict[str, float | int | str | list | dict | bool]) -> list[Workspace]:
    return [
        Workspace(
            id=e.get("id"),
            name=e.get("name"),
            active=e.get("active"),
            size_allowance=int(e.get("sizeAllowance")),
            current_size=int(e.get("currentSize")),
        )
        for e in response.get("workspaces")
    ]


def to_models(response: dict[str, float | int | str | list | dict | bool]) -> list[Model]:
    return [
        Model(
            id=e.get("id"),
            name=e.get("name"),
            active_state=e.get("activeState"),
            last_saved_serial_number=int(e.get("lastSavedSerialNumber")),
            last_modified_by_user_guid=e.get("lastModifiedByUserGuid"),
            memory_usage=int(e.get("memoryUsage", 0)),
            current_workspace_id=e.get("currentWorkspaceId"),
            current_workspace_name=e.get("currentWorkspaceName"),
            model_url=e.get("modelUrl"),
            category_values=e.get("categoryValues"),
            iso_creation_date=e.get("isoCreationDate"),
            last_modified=e.get("lastModified"),
        )
        for e in response.get("models")
    ]


def determine_action_type(action_id: int) -> str:
    if 12000000000 <= action_id < 113000000000:
        return "imports"
    if 116000000000 <= action_id < 117000000000:
        return "exports"
    if 117000000000 <= action_id < 118000000000:
        return "actions"
    if 118000000000 <= action_id < 119000000000:
        return "processes"
    raise InvalidIdentifierException(f"Action '{action_id}' is not a valid identifier.")



from ._async_client import AsyncClient
from ._client import Client
from ._exceptions import (
    AnaplanException,
    InvalidIdentifierException,
    InvalidCredentialsException,
    AnaplanActionError,
)

__all__ = [
    "AsyncClient",
    "Client",
    "AnaplanException",
    "InvalidCredentialsException",
    "InvalidIdentifierException",
    "AnaplanActionError",
]
